from .attributes import BaseAttribute, GenericFieldAttribute, \
    TagFieldAttribute, TimestampFieldAttribute
from .db.query import Query, BulkInsertQuery
from .response import InfluxDBResponse
from .serializers import MeasurementPointSerializer


class MeasurementMeta(type):
    def __init__(cls, name, *args, **kwargs):
        super(MeasurementMeta, cls).__init__(name, *args, **kwargs)
        attribute_names = cls._get_attribute_names()
        cls._extend_attributes(attribute_names)

        get_query = cls._factory_get_query()
        setattr(cls, 'get_query', get_query)

    def __call__(cls, *args, **kwargs):
        setattr(cls, '_get_attributes', cls._get_attributes)
        instance = type.__call__(cls, *args, **kwargs)
        return instance

    def _factory_get_query(cls):
        def get_query():
            class MeasurementQuery(Query):
                def format(self, result, parser_class=cls.parser_class):
                    return parser_class(result, cls).convert()

                def evaluate(self, parser_class=cls.parser_class):
                    result = InfluxDBResponse(self.execute())
                    formatted_result = self.format(result, parser_class)
                    return formatted_result
            return MeasurementQuery().from_measurements(cls.measurement_name)
        return get_query

    def _get_attribute_names(cls):
        def filter_func(x):
            return isinstance(variables[x], BaseAttribute)
        variables = cls.__dict__
        attribute_names = list(filter(filter_func, variables))
        return attribute_names

    def _get_attributes(cls):
        def filter_func(x):
            return isinstance(x, BaseAttribute)
        variables = cls.__dict__.values()
        attributes = list(filter(filter_func, variables))
        return attributes

    def _get_timestamp_attributes(cls):
        def filter_func(x):
            return isinstance(x, TimestampFieldAttribute)
        attributes = cls._get_attributes()
        timestamp_attributes = list(filter(filter_func, attributes))
        return timestamp_attributes

    def _extend_attributes(cls, attribute_names):
        def generate_getter_and_setter(attr_name):
            def getx(self):
                attribute_field = getattr(self, attr_name)
                return attribute_field.get_internal_value()

            def setx(self, value):
                attribute_field = getattr(self, attr_name)
                attribute_field.set_internal_value(value)
            return getx, setx

        for attribute_name in attribute_names:
            ext_attribute_name = '_extended_' + attribute_name
            attribute_field = getattr(cls, attribute_name)
            attribute_field.attribute_name = attribute_name
            attribute_field.ext_attribute_name = ext_attribute_name

            getx, setx = generate_getter_and_setter(ext_attribute_name)
            prop = property(getx, setx)
            setattr(cls, ext_attribute_name, attribute_field)
            setattr(cls, attribute_name, prop)


class Measurement(object, metaclass=MeasurementMeta):
    parser_class = MeasurementPointSerializer
    measurement_name = 'default'

    def __init__(self, **kwargs):
        self.check_attribute_values(**kwargs)
        self.clone_attributes()
        for key, value in kwargs.items():
            setattr(self, key, value)

    def check_attribute_values(self, **kwargs):
        def filter_required_attributes(x):
            return not x.default and not x.is_nullable

        attributes = self._get_attributes()
        required_attributes = list(filter(
            filter_required_attributes,
            attributes,
        ))
        required_attributes_names = [
            r.attribute_name
            for r in required_attributes
        ]
        for key in required_attributes_names:
            if key not in kwargs:
                raise AttributeError('The field {} is not defined'.format(key))

    def clone_attributes(self):
        attributes = self._get_attributes()
        for attr in attributes:
            cloned_attribute = attr.clone()
            setattr(self, attr.ext_attribute_name, cloned_attribute)

    def dict(self):
        dict_values = {}
        attributes = self.get_attributes()
        for attr in attributes:
            dict_values[attr.attribute_name] = getattr(self, attr.attribute_name)
        return dict_values

    def get_attributes(self):
        def filter_func(x):
            return isinstance(x, BaseAttribute)
        variables = self.__dict__.values()
        attributes = list(filter(filter_func, variables))
        return attributes

    def get_attribute_names(self):
        attributes = self.get_attributes()
        attribute_names = [attr.attribute_name for attr in attributes]
        return attribute_names

    def get_ext_attribute_names(self):
        attributes = self.get_attributes()
        attribute_names = [attr.ext_attribute_name for attr in attributes]
        return attribute_names

    def get_timestamp_attributes(self):
        def filter_func(x):
            return isinstance(x, TimestampFieldAttribute)
        attributes = self.get_attributes()
        timestamp_attributes = list(filter(filter_func, attributes))
        return timestamp_attributes

    def get_prep_value(self):
        def factory_filter_func(cls):
            def filter_func(attr):
                return isinstance(attr, cls)
            return filter_func

        attributes = self.get_attributes()
        field_attributes = list(filter(
            factory_filter_func(GenericFieldAttribute),
            attributes,
        ))
        tag_attributes = list(filter(
            factory_filter_func(TagFieldAttribute),
            attributes,
        ))
        timestamp_attributes = list(filter(
            factory_filter_func(TimestampFieldAttribute),
            attributes,
        ))

        prep_value_groups = []
        attributes_groups = [field_attributes, tag_attributes, timestamp_attributes]
        for attr_group in attributes_groups:
            prep_value_group = []
            for attr in attr_group:
                attr_prep_value = attr.get_prep_value()
                attr_name = attr.name or attr.attribute_name
                prep_value = '{}={}'.format(attr_name, attr_prep_value)
                prep_value_group.append(prep_value)
            str_prep_value_group = ','.join(prep_value_group)
            prep_value_groups.append(str_prep_value_group)

        prep_value_groups[0] = ','.join([self.measurement_name] + [prep_value_groups[0]])
        final_prep_value = ' '.join(prep_value_groups)
        return final_prep_value

    def items(self):
        return self.dict().items()

    @staticmethod
    def bulk_save(points):
        str_points = '\n'.join([point.get_prep_value() for point in points])
        return BulkInsertQuery(str_points).execute()



from . import _generate_function

Abs = _generate_function('ABS')
ACos = _generate_function('ACOS')
ASin = _generate_function('ASIN')
ATan = _generate_function('ATAN')
ATan2 = _generate_function('ATAN2')
Ceil = _generate_function('CEIL')
Cos = _generate_function('COS')
CumulativeSum = _generate_function('CUMULATIVE_SUM')
Derivative = _generate_function('DERIVATIVE')
Difference = _generate_function('DIFFERENCE')
Elapsed = _generate_function('ELAPSED')
Exp = _generate_function('EXP')
Floor = _generate_function('FLOOR')
Histogram = _generate_function('HISTOGRAM')
Ln = _generate_function('LN')
Log = _generate_function('LOG')
Log2 = _generate_function('LOG2')
Log10 = _generate_function('LOG10')
MovingAverage = _generate_function('MOVING_AVERAGE')
NonNegativeDerivative = _generate_function('NON_NEGATIVE_DERIVATIVE')
NonNegativeDifference = _generate_function('NON_NEGATIVE_DIFFERENCE')
Pow = _generate_function('POW')
Round = _generate_function('ROUND')
Sin = _generate_function('SIN')
Sqrt = _generate_function('SQRT')
Tan = _generate_function('TAN')



from enum import Enum


class WhereOperator(Enum):
    LT = 'lt'
    LTE = 'lte'
    GT = 'gt'
    GTE = 'gte'
    EQ = 'eq'
    NE = 'ne'


EVALUATED_OPERATORS = {
    WhereOperator.LT: '<',
    WhereOperator.LTE: '<=',
    WhereOperator.GT: '>',
    WhereOperator.GTE: '>=',
    WhereOperator.EQ: '=',
    WhereOperator.NE: '!=',
}

INVERTED_OPERATORS = {
    WhereOperator.LT: WhereOperator.GTE,
    WhereOperator.LTE: WhereOperator.GT,
    WhereOperator.GT: WhereOperator.LTE,
    WhereOperator.GTE: WhereOperator.LT,
    WhereOperator.EQ: WhereOperator.NE,
    WhereOperator.NE: WhereOperator.EQ,
}


class Field:
    def __init__(self, field_name):
        self.field_name = field_name

    def __lt__(self, value):
        return Criteria(self, value, WhereOperator.LT)

    def __le__(self, value):
        return Criteria(self, value, WhereOperator.LTE)

    def __eq__(self, value):
        return Criteria(self, value, WhereOperator.EQ)

    def __ne__(self, value):
        return Criteria(self, value, WhereOperator.NE)

    def __ge__(self, value):
        return Criteria(self, value, WhereOperator.GTE)

    def __gt__(self, value):
        return Criteria(self, value, WhereOperator.GT)

    def __str__(self):
        return self.field_name


class Criteria:
    def __init__(self, left_operand, right_operand, operator):
        self.left_operand = left_operand
        self.right_operand = right_operand
        self.operator = operator

    def __invert__(self):
        inverted_operator = INVERTED_OPERATORS[self.operator]
        return Criteria(
            self.left_operand,
            self.right_operand,
            inverted_operator,
        )

    def __or__(self, criteria):
        return DisjunctionCriteria(self, criteria)

    def evaluate(self):
        left_operand = '"{}"'.format(self.left_operand)
        operator = EVALUATED_OPERATORS[self.operator]
        right_operand = self.right_operand
        if isinstance(right_operand, str):
            right_operand = '\'{}\''.format(self.right_operand)
        return '{} {} {}'.format(left_operand, operator, right_operand)

    def __str__(self):
        return 'CRITERIA: {} {} {}'.format(
            self.left_operand,
            EVALUATED_OPERATORS[self.operator],
            self.right_operand,
        )


class DisjunctionCriteria:
    def __init__(self, left_criteria, right_criteria):
        self.left_criteria = left_criteria
        self.right_criteria = right_criteria

    def __or__(self, criteria):
        return DisjunctionCriteria(self, criteria)

    def evaluate(self):
        left_criteria = '({}'.format(self.left_criteria.evaluate())
        right_criteria = '{})'.format(self.right_criteria.evaluate())
        return ' OR '.join([left_criteria, right_criteria])



from functools import lru_cache
from .function import aggregations
from ..response import InfluxDBResponse
from ..serializers import BaseSerializer
from .. import Influxable


class RawQuery:
    def __init__(self, str_query):
        self.str_query = str_query

    def execute(self):
        return self.raw_response

    @property
    def raw_response(self):
        return self._resolve()

    @lru_cache(maxsize=None)
    def _resolve(self, *args, **kwargs):
        instance = Influxable.get_instance()
        return instance.execute_query(query=self.str_query, method='post')


class Query(RawQuery):
    def __init__(self):
        self.initial_query = '{select_clause} {from_clause}'
        self.from_clause = 'FROM {measurements}'
        self.select_clause = 'SELECT {fields}'
        self.where_clause = ' WHERE {criteria}'
        self.selected_fields = '*'
        self.selected_criteria = []
        self.selected_measurements = 'default'
        self.limit_value = None
        self.slimit_value = None
        self.offset_value = None
        self.soffset_value = None

    def from_measurements(self, *measurements):
        quoted_measurements = ['"{}"'.format(m) for m in measurements]
        self.selected_measurements = ', '.join(quoted_measurements)
        return self

    def select(self, *fields):
        evaluated_fields = []
        for f in fields:
            evaluated_field = f.evaluate() if hasattr(f, 'evaluate') else f
            evaluated_fields.append(evaluated_field)
        self.selected_fields = ', '.join(evaluated_fields)
        return self

    def where(self, *criteria):
        self.selected_criteria = list(criteria)
        return self

    def limit(self, value):
        self.limit_value = value
        return self

    def slimit(self, value):
        self.slimit_value = value
        return self

    def offset(self, value):
        self.offset_value = value
        return self

    def soffset(self, value):
        self.soffset_value = value
        return self

    def count(self, value='*'):
        return self.select(aggregations.Count(value))

    def distinct(self, value='*'):
        return self.select(aggregations.Distinct(value))

    def integral(self, value='*'):
        return self.select(aggregations.Integral(value))

    def mean(self, value='*'):
        return self.select(aggregations.Mean(value))

    def median(self, value='*'):
        return self.select(aggregations.Median(value))

    def mode(self, value='*'):
        return self.select(aggregations.Mode(value))

    def spread(self, value='*'):
        return self.select(aggregations.Spread(value))

    def std_dev(self, value='*'):
        return self.select(aggregations.StdDev(value))

    def sum(self, value='*'):
        return self.select(aggregations.Sum(value))

    def _prepare_query(self):
        select_clause = self.select_clause.format(fields=self.selected_fields)
        from_clause = self.from_clause.format(measurements=self.selected_measurements)
        prepared_query = self.initial_query.format(
            select_clause=select_clause,
            from_clause=from_clause,
        )
        if len(self.selected_criteria):
            criteria = [c.evaluate() for c in self.selected_criteria]
            eval_criteria = ' AND '.join(criteria)
            self.where_clause = self.where_clause.format(criteria=eval_criteria)
            prepared_query += self.where_clause
        if self.limit_value is not None:
            self.limit_clause = ' LIMIT {}'.format(self.limit_value)
            prepared_query += self.limit_clause
        if self.offset_value is not None:
            self.offset_clause = ' OFFSET {}'.format(self.offset_value)
            prepared_query += self.offset_clause
        if self.slimit_value is not None:
            self.slimit_clause = ' SLIMIT {}'.format(self.slimit_value)
            prepared_query += self.slimit_clause
        if self.soffset_value is not None:
            self.soffset_clause = ' SOFFSET {}'.format(self.soffset_value)
            prepared_query += self.soffset_clause
        print('prepared_query', prepared_query)
        return prepared_query

    def execute(self):
        prepared_query = self._prepare_query()
        self.str_query = prepared_query
        return super().execute()

    def format(self, result, parser_class=BaseSerializer, **kwargs):
        return parser_class(result, **kwargs).convert()

    def evaluate(self, parser_class=BaseSerializer, **kwargs):
        result = InfluxDBResponse(self.execute())
        formatted_result = self.format(result, parser_class, **kwargs)
        return formatted_result


class BulkInsertQuery(RawQuery):
    @lru_cache(maxsize=None)
    def _resolve(self, *args, **kwargs):
        instance = Influxable.get_instance()
        return instance.write_points(points=self.str_query)



# copied from:
# https://stackoverflow.com/questions/31875/is-there-a-simple-elegant-way-to-define-singletons


class Singleton:
    """
    A non-thread-safe helper class to ease implementing singletons.
    This should be used as a decorator -- not a metaclass -- to the
    class that should be a singleton.
    The decorated class can define one `__init__` function that
    takes only the `self` argument. Also, the decorated class cannot be
    inherited from. Other than that, there are no restrictions that apply
    to the decorated class.
    To get the singleton instance, use the `instance` method. Trying
    to use `__call__` will result in a `TypeError` being raised.
    """

    def __init__(self, decorated):
        self._decorated = decorated

    def get_instance(self, *args, **kwargs):
        """
        Returns the singleton instance. Upon its first call, it creates a
        new instance of the decorated class and calls its `__init__` method.
        On all subsequent calls, the already created instance is returned.
        """
        try:
            return self._instance
        except AttributeError:
            self._instance = self._decorated()
            return self._instance

    def __call__(self, *args, **kwargs):
        if hasattr(self, '_instance'):
            msg = 'Singletons must be accessed through `get_instance()`'
            raise TypeError(msg)
        else:
            self._instance = self._decorated(*args, **kwargs)
            return self._instance

    def __instancecheck__(self, inst):
        return isinstance(inst, self._decorated)



from . import _generate_function

Count = _generate_function('COUNT')
Distinct = _generate_function('DISTINCT')
Integral = _generate_function('INTEGRAL')
Mean = _generate_function('MEAN')
Median = _generate_function('MEDIAN')
Mode = _generate_function('MODE')
Spread = _generate_function('SPREAD')
StdDev = _generate_function('STDDEV')
Sum = _generate_function('SUM')



from . import _generate_function, _generate_function_with_param

Bottom = _generate_function_with_param('BOTTOM')
First = _generate_function('FIRST')
Last = _generate_function('LAST')
Max = _generate_function('MAX')
Min = _generate_function('MIN')
Percentile = _generate_function_with_param('PERCENTILE')
Sample = _generate_function_with_param('SAMPLE')
Top = _generate_function_with_param('TOP')



class InfluxDBResponse:
    def __init__(self, raw_json):
        self._raw_json = raw_json

    @property
    def raw(self):
        return self._raw_json

    @property
    def main_serie(self):
        series = self.series
        if len(series):
            return series[0]
        return None

    @property
    def series(self):
        if 'results' in self.raw:
            results = self.raw['results']
            if len(results):
                result = results[0]
                if 'series' in result:
                    return [InfluxDBSerieResponse(s) for s in result['series']]
        return []


class InfluxDBSerieResponse:
    def __init__(self, json_serie):
        self._raw_json_serie = json_serie

    @property
    def raw(self):
        return self._raw_json_serie

    @property
    def columns(self):
        return self._raw_json_serie["columns"]

    @property
    def name(self):
        return self._raw_json_serie.get("name", "default")

    @property
    def values(self):
        return self._raw_json_serie["values"]


class InfluxDBErrorResponse:
    def __init__(self, raw_json):
        self._raw_json = raw_json

    @property
    def raw(self):
        return self._raw_json

    @property
    def error(self):
        return self.raw['error']



import json
import itertools
import pandas as pd


class BaseSerializer:
    def __init__(self, response):
        self.response = response

    def convert(self):
        return self.response


class JsonSerializer(BaseSerializer):
    def convert(self):
        return json.dumps(self.response.raw)


class FormattedSerieSerializer(BaseSerializer):
    def convert(self):
        formatted_series = []
        series = self.response.series
        for serie in series:
            name = serie.name
            columns = serie.columns
            values = serie.values
            formatted_values = [dict(zip(columns, v)) for v in values]
            formatted_series.append({name: formatted_values})
        return formatted_series


class FlatFormattedSerieSerializer(FormattedSerieSerializer):
    def convert(self):
        formatted_series = super().convert()
        if len(formatted_series) == 1:
            main_serie = formatted_series[0]
            flat_main_serie = list(main_serie.values())[0]
            return flat_main_serie
        return []


class FlatSimpleResultSerializer(BaseSerializer):
    def convert(self):
        serie = self.response.main_serie
        values = serie.values
        flatten_serie = list(itertools.chain(*values))
        return flatten_serie


class PandasSerializer(BaseSerializer):
    def convert(self):
        serie = self.response.main_serie
        columns = serie.columns
        values = serie.values
        df = pd.DataFrame(values, columns=columns)
        return df


class MeasurementPointSerializer(FlatFormattedSerieSerializer):
    def __init__(self, response, measurement):
        self.response = response
        self.measurement = measurement

    def convert(self):
        flat_formatted_series = super().convert()
        timestamp_attributes = self.measurement._get_timestamp_attributes()
        timestamp_attributes_names = [
            ta.attribute_name
            for ta in timestamp_attributes
        ]
        self.convert_to_seconds(timestamp_attributes_names, flat_formatted_series)
        points = [self.measurement(**ffs) for ffs in flat_formatted_series]
        return points

    def convert_to_seconds(self, attr_names, series):
        NANO_TO_SEC_RATIO = 1000 * 1000 * 1000
        for field in series:
            for attr_name in attr_names:
                field[attr_name] /= NANO_TO_SEC_RATIO



from . import settings
from .request import InfluxDBRequest


class Connection:
    def __init__(self, *args, **kwargs):
        self.base_url = kwargs.get('base_url', settings.INFLUXDB_URL)
        self.user = kwargs.get('user', settings.INFLUXDB_USER)
        self.password = kwargs.get('password', settings.INFLUXDB_PASSWORD)
        self.database_name = kwargs.get(
            'database_name',
            settings.INFLUXDB_DATABASE_NAME,
        )

        self.request = InfluxDBRequest(self.base_url, self.database_name)
        self.request.auth = (self.user, self.password)

    @staticmethod
    def create(base_url, database_name, user='', password=''):
        return Connection(base_url, database_name, user, password)



from .query import RawQuery
from ..response import InfluxDBResponse
from .. import serializers


class InfluxDBAdmin:
    @staticmethod
    def _execute_query(query, parser=serializers.FlatFormattedSerieSerializer):
        response = RawQuery(query).execute()
        influx_response = InfluxDBResponse(response)
        formatted_result = parser(influx_response).convert()
        return formatted_result

    @staticmethod
    def show_databases():
        query = 'SHOW DATABASES'
        parser = serializers.FlatSimpleResultSerializer
        return InfluxDBAdmin._execute_query(query, parser)

    @staticmethod
    def show_measurements():
        query = 'SHOW MEASUREMENTS'
        parser = serializers.FlatSimpleResultSerializer
        return InfluxDBAdmin._execute_query(query, parser)

    @staticmethod
    def show_queries():
        query = 'SHOW QUERIES'
        parser = serializers.FlatFormattedSerieSerializer
        return InfluxDBAdmin._execute_query(query, parser)

    @staticmethod
    def show_series():
        query = 'SHOW SERIES'
        parser = serializers.FlatFormattedSerieSerializer
        return InfluxDBAdmin._execute_query(query, parser)

    @staticmethod
    def show_stats():
        query = 'SHOW STATS'
        parser = serializers.FormattedSerieSerializer
        return InfluxDBAdmin._execute_query(query, parser)

    @staticmethod
    def show_users():
        query = 'SHOW USERS'
        parser = serializers.FlatFormattedSerieSerializer
        return InfluxDBAdmin._execute_query(query, parser)



def _generate_function(identifier):
    class Function:
        def __init__(self, field='*'):
            self.field = field

        def evaluate(self):
            if hasattr(self.field, 'evaluate'):
                field = self.field.evaluate()
            else:
                field = self.field
            return '{}({})'.format(identifier, field)
    return Function


def _generate_function_with_param(identifier):
    class Function:
        def __init__(self, n, *fields):
            self.n = n
            self.fields = fields

        def evaluate(self):
            fields = ', '.join(self.fields)
            return '{}({}, {})'.format(identifier, fields, self.n)
    return Function



class InfluxDBApi:
    @staticmethod
    def get_debug_requests(request, seconds=10):
        url = '/debug/requests'
        params = {'seconds': seconds}
        res = request.get(url=url, params=params)
        return res.json()

    @staticmethod
    def get_debug_vars(request):
        url = '/debug/vars'
        res = request.get(url=url)
        return res.json()

    @staticmethod
    def ping(request, verbose=False):
        url = '/ping'
        params = {'verbose': verbose}
        res = request.get(url=url, params=params)
        return res.text or True

    @staticmethod
    def execute_query(
        request,
        query,
        method='get',
        chunked=False,
        epoch='ns',
        pretty=False,
    ):
        url = '/query'
        params = {
            'db': request.database_name,
            'q': query,
            'epoch': epoch,
            'chunked': chunked,
            'pretty': pretty,
        }
        res = request.request(method, url, params=params)
        return res.json()

    @staticmethod
    def write_points(
        request,
        points,
        precision='ns',
        consistency='all',
        retention_policy_name='DEFAULT',
    ):
        url = '/write'
        params = {
            'db': request.database_name,
            'precision': precision,
            'consistency': consistency,
            'retention_policy_name': retention_policy_name,
        }
        res = request.post(url, params=params, data=points)
        return res.text



from setuptools import find_packages, setup

VERSION = '1.0.1'

with open('requirements.txt', 'r') as f:
    requirements = [x.strip() for x in f if x.strip()]

with open('README.rst', 'r') as f:
    readme = f.read()

setup(
    name='influxable',
    packages=find_packages(),
    version=VERSION,
    license='MIT',
    description='A lightweight python ORM / ODM for InfluxDB',
    long_description=readme,
    author='Javid Mougamadou',
    author_email='javidjms0@gmail.com',
    url='https://github.com/Javidjms/influxable',
    download_url='https://github.com/Javidjms/influxable/archive/1.0.1.zip',
    keywords=['python', 'influxdb', 'odm', 'orm', 'driver', 'client'],
    install_requires=requirements,
    python_requires='>=3.0.*',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    project_urls={
        'Source': 'https://github.com/Javidjms/influxable',
    },
)



import arrow
from decimal import Decimal as D
from .helpers.utils import inv


class TimestampPrecision:
    HOURS = 'h'
    MICROSECONDS = 'u'
    MILLISECONDS = 'ms'
    MINUTES = 'm'
    NANOSECONDS = 'ns'
    SECONDS = 's'


TIMESTAMP_CONVERT_RATIO = {
    TimestampPrecision.HOURS: inv(60 * 60),
    TimestampPrecision.MICROSECONDS: 1 * 1000 * 1000,
    TimestampPrecision.MILLISECONDS: 1000,
    TimestampPrecision.MINUTES: inv(60),
    TimestampPrecision.NANOSECONDS: 1 * 1000 * 1000 * 1000,
    TimestampPrecision.SECONDS: 1,
}


class BaseAttribute:
    def __init__(self, **kwargs):
        self._value = None
        self.raw_value = None
        self.attribute_name = kwargs.get('name', None)
        self.default = kwargs.get('default', None)
        self.enforce_cast = kwargs.get('enforce_cast', True)
        self.is_nullable = kwargs.get('is_nullable', True)

    def clean(self, value):
        if value is None and self.default is not None:
            self._value = self.default
        elif value is None:
            self._value = None

    def clone(self):
        cls = self.__class__
        instance = cls(**self.__dict__)
        if self._value is not None:
            instance.set_internal_value(self._value)
        return instance

    def get_internal_value(self):
        return self._value

    def get_prep_value(self):
        prep_value = self.to_influx(self._value)
        return prep_value

    @property
    def name(self):
        return self.attribute_name

    def to_influx(self, value):
        return str(value)

    def to_python(self, value):
        return str(value)

    def reset(self):
        self._value = None

    def set_internal_value(self, value):
        self.validate(value)
        self.raw_value = value
        if value is not None:
            try:
                self._value = self.to_python(value)
            except ValueError as exception:
                if self.enforce_cast:
                    raise exception
                self._value = value
        self.clean(value)

    def validate(self, value):
        if value is None and self.default is None and not self.is_nullable:
            raise ValueError('This field cannot be nullable')

    @property
    def value(self):
        return self.get_internal_value()


class GenericFieldAttribute(BaseAttribute):
    pass


class IntegerFieldAttribute(GenericFieldAttribute):
    def __init__(self, **kwargs):
        super(IntegerFieldAttribute, self).__init__(**kwargs)
        self.min_value = kwargs.get('min_value', None)
        self.max_value = kwargs.get('max_value', None)

    def to_influx(self, value):
        str_value = str(value)
        return "{}i".format(str_value)

    def to_python(self, value):
        return int(value)

    def validate(self, value):
        super(IntegerFieldAttribute, self).validate(value)
        if value is not None and self.min_value is not None and int(value) < self.min_value:
            raise ValueError('The value must be greater than the min_value')
        if value is not None and self.max_value is not None and int(value) > self.max_value:
            raise ValueError('The value must be lower than the max_value')


class FloatFieldAttribute(IntegerFieldAttribute):
    def __init__(self, **kwargs):
        super(FloatFieldAttribute, self).__init__(**kwargs)
        self.max_nb_decimals = kwargs.get('max_nb_decimals', None)

    def clean(self, value):
        super(FloatFieldAttribute, self).clean(value)
        if self.max_nb_decimals is not None:
            precision = '.' + '0' * (self.max_nb_decimals - 1) + '1'
            self._value = self.to_python(value).quantize(D(precision))

    def to_influx(self, value):
        str_value = str(value)
        return str_value

    def to_python(self, value):
        return D(value)

    def validate(self, value):
        super(FloatFieldAttribute, self).validate(value)
        if self.max_nb_decimals is not None and not isinstance(self.max_nb_decimals, int):
            raise ValueError('max_nb_decimals must be integer')
        if self.max_nb_decimals is not None and self.max_nb_decimals <= 0:
            raise ValueError('max_nb_decimals must be positive')


class StringFieldAttribute(GenericFieldAttribute):
    def __init__(self, **kwargs):
        super(StringFieldAttribute, self).__init__(**kwargs)
        self.choices = kwargs.get('choices', None)
        self.max_length = kwargs.get('max_length', None)

    def to_influx(self, value):
        str_value = str(value)
        return "\'{}\'".format(str_value)

    def to_python(self, value):
        return str(value)

    def validate(self, value):
        super(StringFieldAttribute, self).validate(value)
        if self.choices is not None and not isinstance(self.choices, list):
            raise ValueError('choices must be a list')
        if self.choices is not None and not any(
            [isinstance(c, str) for c in self.choices]
        ):
            raise ValueError('choices items must be a string')
        if self.choices is not None and value not in self.choices:
            raise ValueError('The value is not refered in choices')
        if self.max_length is not None and not isinstance(self.max_length, int):
            raise ValueError('max_length must be integer')
        if self.max_length is not None and self.max_length <= 0:
            raise ValueError('max_length must be positive')
        if self.max_length is not None and len(str(value)) > self.max_length:
            raise ValueError('the string length must be lower than the max_length')


class BooleanFieldAttribute(GenericFieldAttribute):
    def to_influx(self, value):
        str_value = str(value).lower()
        return str_value

    def to_python(self, value):
        return bool(value)


class TagFieldAttribute(BaseAttribute):
    pass


class TimestampFieldAttribute(BaseAttribute):
    def __init__(self, **kwargs):
        super(TimestampFieldAttribute, self).__init__(**kwargs)
        self.auto_now = kwargs.get('auto_now', True)
        self.precision = kwargs.get('precision', 'ns')

    def clean(self, value):
        super(TimestampFieldAttribute, self).clean(value)
        if value is None and self.auto_now:
            timestamp = arrow.now().timestamp
            self._value = self.to_python(timestamp)
            self.formatted_timestamp = self.convert_to_nanoseconds(timestamp)
        elif value:
            self.formatted_timestamp = self.convert_to_nanoseconds(value)
        else:
            self.formatted_timestamp = None

    def convert_to_nanoseconds(self, timestamp):
        precision = TimestampPrecision.NANOSECONDS
        return self.convert_to_precision(timestamp, precision)

    def convert_to_precision(self, timestamp, precision):
        convert_ratio = TIMESTAMP_CONVERT_RATIO[precision]
        return D(timestamp) * convert_ratio

    def to_influx(self, value):
        timestamp = D(self.formatted_timestamp)
        str_value = str(timestamp)
        return "{}".format(str_value)

    def to_python(self, value):
        timestamp = D(value)
        return self.convert_to_precision(timestamp, self.precision)

    def validate(self, value):
        super(TimestampFieldAttribute, self).validate(value)
        if self.precision not in TIMESTAMP_CONVERT_RATIO:
            raise ValueError('precision must be one of [ns,u,ms,s,m,h]')


class DateTimeFieldAttribute(TimestampFieldAttribute):
    def __init__(self, **kwargs):
        super(DateTimeFieldAttribute, self).__init__(**kwargs)
        self.str_format = kwargs.get('str_format', 'YYYY-MM-DD HH:mm:ss ZZ')

    def clean(self, value):
        super(DateTimeFieldAttribute, self).clean(value)
        if value is None and self.auto_now:
            timestamp = arrow.now().datetime
            self._value = self.to_python(timestamp)

    def get_internal_value(self):
        return arrow.get(self._value).format(self.str_format)

    def to_influx(self, value):
        timestamp = arrow.get(value).timestamp
        nanoseconds = self.convert_to_nanoseconds(timestamp)
        str_value = str(nanoseconds)
        return "{}".format(str_value)

    def to_python(self, value):
        datetime = arrow.get(value).datetime
        return datetime



import os

INFLUXDB_URL = os.getenv('INFLUXDB_URL', 'http://localhost:8086')
INFLUXDB_USER = os.getenv('INFLUXDB_USER', 'admin')
INFLUXDB_PASSWORD = os.getenv('INFLUXDB_PASSWORD', 'changeme')
INFLUXDB_DATABASE_NAME = os.getenv('INFLUXDB_DATABASE_NAME', 'default')
